@@ -1,5 +1,5 @@
 /*
- * "$Id: file.c,v 1.13.2.13 2001/03/04 03:05:05 mike Exp $"
+ * "$Id: file.c,v 1.13.2.25 2001/10/20 21:49:16 mike Exp $"
  *
  *   Filename routines for HTMLDOC, a HTML document processing program.
  *
@@ -43,15 +43,46 @@
 #include "http.h"
 #include "progress.h"
 
-#if defined(WIN32) || defined(__EMX__)
+#if defined(WIN32)
 #  include <io.h>
 #else
 #  include <unistd.h>
-#endif /* WIN32 || __EMX__ */
+#endif /* WIN32 */
 
 #include <errno.h>
 #include <fcntl.h>
+#include <sys/types.h>
 #include <sys/stat.h>
+#include <ctype.h>
+
+
+
+/*
+ * Temporary file definitions...
+ */
+
+#ifdef WIN32
+#  define getpid	GetCurrentProcessId
+#  define TEMPLATE	"%s%08x.%06d.tmp"
+#  define OPENMODE	(_O_CREAT | _O_RDWR | _O_EXCL | \
+			 _O_TRUNC | _O_BINARY | _O_SHORT_LIVED)
+#  define OPENPERM	(_S_IREAD | _S_IWRITE)
+#else
+#  define TEMPLATE	"%s/%06d.%06d.tmp"
+#  define OPENMODE	(O_CREAT | O_RDWR | O_EXCL | O_TRUNC)
+#  define OPENPERM	0600
+#endif /* WIN32 */
+
+
+/*
+ * Cache file structure...
+ */
+
+typedef struct		/* Cache for all temporary files */
+{
+  char	*name;				/* Temporary filename */
+  char	*url;				/* URL */
+} cache_t;
 
 
 /*
@@ -61,12 +92,9 @@
 char	proxy_host[HTTP_MAX_URI] = "";	/* Proxy hostname */
 int	proxy_port = 0;			/* Proxy port */
 http_t	*http = NULL;			/* Connection to remote server */
-int	web_files = 0;			/* Number of temporary files */
-struct					/* Cache for all temporary files */
-{
-  char	*name;				/* Temporary filename */
-  char	*url;				/* URL */
-}	web_cache[2 * MAX_IMAGES];
+int	web_files = 0,			/* Number of temporary files */
+	web_alloc = 0;			/* Number of allocated files */
+cache_t	*web_cache = NULL;		/* Cache array */
 
 
 /*
@@ -101,7 +129,7 @@ file_basename(const char *s)	/* I - Filename or URL */
     return (basename);
 
   strcpy(buf, basename);
-  *strchr(buf, '#') = '\0';
+  *(char *)strchr(buf, '#') = '\0';
 
   return (buf);
 }
@@ -137,14 +165,14 @@ file_cleanup(void)
 
   while (web_files > 0)
   {
-#ifdef WIN32
-    snprintf(filename, sizeof(filename), "%s/%06d.%06d.dat", tmpdir,
-             GetCurrentProcessId(), web_files);
-#else
-    snprintf(filename, sizeof(filename), "%s/%06d.%06d", tmpdir, getpid(), web_files);
-#endif /* WIN32 */
+    snprintf(filename, sizeof(filename), TEMPLATE, tmpdir,
+             getpid(), web_files);
+
+    if (unlink(filename))
+      progress_error(HD_ERROR_DELETE_ERROR,
+                     "Unable to delete temporary file \"%s\": %s",
+                     filename, strerror(errno));
 
-    unlink(filename);
     web_files --;
 
     if (web_cache[web_files].name)
@@ -152,6 +180,14 @@ file_cleanup(void)
     if (web_cache[web_files].url)
       free(web_cache[web_files].url);
   }
+
+  if (web_alloc)
+  {
+    free(web_cache);
+
+    web_alloc = 0;
+    web_cache = NULL;
+  }
 }
 
 
@@ -169,7 +205,7 @@ file_directory(const char *s)	/* I - Filename or URL */
   if (s == NULL)
     return (NULL);
 
-  if (strncmp(s, "http://", 7) == 0)
+  if (strncmp(s, "http://", 7) == 0 || strncmp(s, "https://", 8) == 0)
   {
    /*
     * Handle URLs...
@@ -254,7 +290,7 @@ file_extension(const char *s)	/* I - Filename or URL */
     return (extension);
 
   strcpy(buf, extension);
-  *strchr(buf, '#') = '\0';
+  *(char *)strchr(buf, '#') = '\0';
 
   return (buf);
 }
@@ -295,9 +331,14 @@ file_find(const char *path,		/* I - Path "dir;dir;dir" */
   if (s == NULL)
     return (NULL);
 
-  if (strncmp(s, "http:", 5) == 0 ||
+  if (strncmp(s, "http:", 5) == 0 || strncmp(s, "//", 2) == 0 ||
       (path != NULL && strncmp(path, "http:", 5) == 0))
     strcpy(method, "http");
+#ifdef HAVE_LIBSSL
+  else if (strncmp(s, "https:", 6) == 0 ||
+           (path != NULL && strncmp(path, "https:", 6) == 0))
+    strcpy(method, "https");
+#endif /* HAVE_LIBSSL */
   else
     strcpy(method, "file");
 
@@ -350,6 +391,13 @@ file_find(const char *path,		/* I - Path "dir;dir;dir" */
       if (!access(filename, 0))
 	return (filename);
     }
+
+   /*
+    * If the filename exists, return the filename...
+    */
+
+    if (!access(s, 0))
+      return ((char *)s);
   }
   else
   {
@@ -362,7 +410,12 @@ file_find(const char *path,		/* I - Path "dir;dir;dir" */
       if (web_cache[i].url && strcmp(web_cache[i].url, s) == 0)
         return (web_cache[i].name);
 
-    if (strncmp(s, "http:", 5) == 0)
+#ifdef HAVE_LIBSSL
+    if (strncmp(s, "http:", 5) == 0 || strncmp(s, "//", 2) == 0 ||
+        strncmp(s, "https:", 6) == 0)
+#else
+    if (strncmp(s, "http:", 5) == 0 || strncmp(s, "//", 2) == 0)
+#endif /* HAVE_LIBSSL */
       httpSeparate(s, method, username, hostname, &port, resource);
     else if (s[0] == '/')
     {
@@ -415,10 +468,21 @@ file_find(const char *path,		/* I - Path "dir;dir;dir" */
       {
         progress_show("Connecting to %s...", connhost);
         atexit(file_cleanup);
-        if ((http = httpConnect(connhost, connport)) == NULL)
-        {
+
+#ifdef HAVE_LIBSSL
+        if (strcmp(method, "http") == 0)
+          http = httpConnect(connhost, connport);
+	else
+          http = httpConnectEncrypt(connhost, connport, HTTP_ENCRYPT_ALWAYS);
+#else
+        http = httpConnect(connhost, connport);
+#endif /* HAVE_LIBSSL */
+
+        if (http == NULL)
+	{
           progress_hide();
-          progress_error("Unable to connect to %s!", connhost);
+          progress_error(HD_ERROR_NETWORK_ERROR,
+	                 "Unable to connect to %s!", connhost);
           return (NULL);
         }
       }
@@ -463,7 +527,7 @@ file_find(const char *path,		/* I - Path "dir;dir;dir" */
     if (status != HTTP_OK)
     {
       progress_hide();
-      progress_error("HTTP error %d: %s!", status, httpStatus(status));
+      progress_error((HDerror)status, "%s", httpStatus(status));
       httpFlush(http);
       return (NULL);
     }
@@ -471,7 +535,9 @@ file_find(const char *path,		/* I - Path "dir;dir;dir" */
     if ((fp = file_temp(filename, sizeof(filename))) == NULL)
     {
       progress_hide();
-      progress_error("Unable to create temporary file \"%s\"!", filename);
+      progress_error(HD_ERROR_WRITE_ERROR,
+                     "Unable to create temporary file \"%s\": %s", filename,
+                     strerror(errno));
       httpFlush(http);
       return (NULL);
     }
@@ -682,6 +748,7 @@ FILE *					/* O - Temporary file */
 file_temp(char *name,			/* O - Filename */
           int  len)			/* I - Length of filename buffer */
 {
+  cache_t	*temp;			/* Pointer to cache entry */
   FILE		*fp;			/* File pointer */
   int		fd;			/* File descriptor */
 #ifdef WIN32
@@ -691,41 +758,61 @@ file_temp(char *name,			/* O - Filename */
 #endif /* WIN32 */
 
 
-  web_cache[web_files].name = NULL;
-  web_cache[web_files].url  = NULL;
+ /*
+  * Allocate memory for the file cache as needed...
+  */
+
+  if (web_files >= web_alloc)
+  {
+    web_alloc += ALLOC_FILES;
+    if (web_files == 0)
+      temp = (cache_t *)malloc(sizeof(cache_t) * web_alloc);
+    else
+      temp = (cache_t *)realloc(web_cache, sizeof(cache_t) * web_alloc);
+
+    if (temp == NULL)
+    {
+      progress_error(HD_ERROR_OUT_OF_MEMORY,
+                     "Unable to allocate memory for %d file entries - %s",
+                     web_alloc, strerror(errno));
+      web_alloc -= ALLOC_FILES;
+      return (NULL);
+    }
+
+    web_cache = temp;
+  }
+
+ /*
+  * Clear a new file cache entry...
+  */
+
+  temp = web_cache + web_files;
+
+  temp->name = NULL;
+  temp->url  = NULL;
   web_files ++;
 
 #ifdef WIN32
   GetTempPath(sizeof(tmpdir), tmpdir);
-
-  snprintf(name, len, "%s/%08x.%06d.dat", tmpdir, GetCurrentProcessId(), web_files);
-
-  fd = _open(name, _O_CREAT | _O_WRONLY | _O_EXCL | _O_TRUNC | _O_SHORT_LIVED |
-                   _O_BINARY, _S_IREAD | _S_IWRITE);
 #else
   if ((tmpdir = getenv("TMPDIR")) == NULL)
     tmpdir = "/var/tmp";
-
-  snprintf(name, len, "%s/%06d.%06d", tmpdir, getpid(), web_files);
-
-  fd = open(name, O_CREAT | O_WRONLY | O_EXCL | O_TRUNC, 0600);
 #endif /* WIN32 */
 
-  if (fd >= 0)
-    fp = fdopen(fd, "wb");
+  snprintf(name, len, TEMPLATE, tmpdir, getpid(), web_files);
+
+  if ((fd = open(name, OPENMODE, OPENPERM)) >= 0)
+    fp = fdopen(fd, "w+b");
   else
     fp = NULL;
 
   if (!fp)
-  {
     web_files --;
-    name[0] = '\0';
-  }
 
   return (fp);
 }
 
 
 /*
- * End of "$Id: file.c,v 1.13.2.13 2001/03/04 03:05:05 mike Exp $".
+ * End of "$Id: file.c,v 1.13.2.25 2001/10/20 21:49:16 mike Exp $".
  */
