@@ -1,5 +1,5 @@
 /*
- * "$Id: html.cxx,v 1.17.2.10 2001/03/08 17:58:30 mike Exp $"
+ * "$Id: html.cxx,v 1.17.2.20 2001/10/20 21:49:16 mike Exp $"
  *
  *   HTML exporting functions for HTMLDOC, a HTML document processing program.
  *
@@ -59,14 +59,19 @@ typedef struct
  */
 
 
-static int	num_links;
-static link_t	links[MAX_LINKS];
+static int	num_links = 0,
+		alloc_links = 0;
+static link_t	*links;
 
 
 /*
  * Local functions...
  */
 
+extern "C" {
+typedef int	(*compare_func_t)(const void *, const void *);
+}
+
 static void	write_header(FILE **out, uchar *filename, uchar *title,
 		             uchar *author, uchar *copyright, uchar *docnumber,
 			     tree_t *t);
@@ -106,7 +111,7 @@ html_export(tree_t *document,	/* I - Document to export */
     image_copy(LogoImage, OutputPath);
 
   if (OutputFiles && TitleImage[0] != '\0' && TitlePage &&
-#if defined(WIN32) || defined(__EMX__)
+#ifdef WIN32
       stricmp(file_extension(TitleImage), "htm") != 0 &&
       stricmp(file_extension(TitleImage), "html") != 0 &&
       stricmp(file_extension(TitleImage), "shtml") != 0)
@@ -114,7 +119,7 @@ html_export(tree_t *document,	/* I - Document to export */
       strcmp(file_extension(TitleImage), "htm") != 0 &&
       strcmp(file_extension(TitleImage), "html") != 0 &&
       strcmp(file_extension(TitleImage), "shtml") != 0)
-#endif // WIN32 || __EMX__
+#endif // WIN32
     image_copy(TitleImage, OutputPath);
 
  /*
@@ -130,7 +135,9 @@ html_export(tree_t *document,	/* I - Document to export */
   * Scan for all links in the document, and then update them...
   */
 
-  num_links = 0;
+  num_links   = 0;
+  alloc_links = 0;
+  links       = NULL;
 
   scan_links(document, NULL);
   update_links(document, NULL);
@@ -177,12 +184,23 @@ html_export(tree_t *document,	/* I - Document to export */
     fputs("</BODY>\n", out);
     fputs("</HTML>\n", out);
 
+    progress_error(HD_ERROR_NONE, "BYTES: %ld", ftell(out));
+
     fclose(out);
   }
 
   if (title != NULL)
     free(title);
 
+  if (alloc_links)
+  {
+    free(links);
+
+    num_links   = 0;
+    alloc_links = 0;
+    links       = NULL;
+  }
+
   return (0);
 }
 
@@ -203,7 +221,7 @@ write_header(FILE   **out,	/* IO - Output file */
   char	realname[1024];		/* Real filename */
   char	*basename;		/* Filename without directory */
   int	newfile;		/* Non-zero if this is a new file */
-  static char	*families[] =	/* Typeface names */
+  static const char *families[] =/* Typeface names */
 		{
 		  "monospace",
 		  "serif",
@@ -243,7 +261,8 @@ write_header(FILE   **out,	/* IO - Output file */
 
   if (*out == NULL)
   {
-    progress_error("Unable to create output file \"%s\" - %s.\n",
+    progress_error(HD_ERROR_WRITE_ERROR,
+                   "Unable to create output file \"%s\" - %s.\n",
                    OutputFiles ? realname : OutputPath,
 		   strerror(errno));
     return;
@@ -363,6 +382,8 @@ write_footer(FILE **out,	/* IO - Output file pointer */
     fputs("</BODY>\n", *out);
     fputs("</HTML>\n", *out);
 
+    progress_error(HD_ERROR_NONE, "BYTES: %ld", ftell(*out));
+
     fclose(*out);
     *out = NULL;
   }
@@ -387,7 +408,7 @@ write_title(FILE  *out,		/* I - Output file */
   if (out == NULL)
     return;
 
-#if defined(WIN32) || defined(__EMX__)
+#ifdef WIN32
   if (stricmp(file_extension(TitleImage), "htm") == 0 ||
       stricmp(file_extension(TitleImage), "html") == 0 ||
       stricmp(file_extension(TitleImage), "shtml") == 0)
@@ -395,12 +416,13 @@ write_title(FILE  *out,		/* I - Output file */
   if (strcmp(file_extension(TitleImage), "htm") == 0 ||
       strcmp(file_extension(TitleImage), "html") == 0 ||
       strcmp(file_extension(TitleImage), "shtml") == 0)
-#endif // WIN32 || __EMX__
+#endif // WIN32
   {
     // Write a title page from HTML source...
     if ((fp = fopen(TitleImage, "rb")) == NULL)
     {
-      progress_error("Unable to open title file \"%s\" - %s!",
+      progress_error(HD_ERROR_FILE_NOT_FOUND,
+                     "Unable to open title file \"%s\" - %s!",
                      TitleImage, strerror(errno));
       return;
     }
@@ -550,7 +572,7 @@ write_all(FILE   *out,		/* I - Output file */
 
       default :
 	  if (t->markup == MARKUP_IMG && OutputFiles &&
-              (src = htmlGetVariable(t, (uchar *)"SRC")) != NULL)
+              (src = htmlGetVariable(t, (uchar *)"REALSRC")) != NULL)
 	  {
 	   /*
             * Update local images...
@@ -575,6 +597,10 @@ write_all(FILE   *out,		/* I - Output file */
 	          t->markup == MARKUP_HR)
 		continue;
 
+	      if (strcasecmp((char *)t->vars[i].name, "REALSRC") == 0 &&
+	          t->markup == MARKUP_IMG)
+		continue;
+
 	      if (col > 72 && !t->preformatted)
 	      {
         	putc('\n', out);
@@ -589,6 +615,8 @@ write_all(FILE   *out,		/* I - Output file */
 
 	      if (t->vars[i].value == NULL)
         	col += fprintf(out, "%s", t->vars[i].name);
+	      else if (strchr((char *)t->vars[i].value, '\"') != NULL)
+        	col += fprintf(out, "%s=\'%s\'", t->vars[i].name, t->vars[i].value);
 	      else
         	col += fprintf(out, "%s=\"%s\"", t->vars[i].name, t->vars[i].value);
 	    }
@@ -617,22 +645,29 @@ write_all(FILE   *out,		/* I - Output file */
 
       switch (t->markup)
       {
-	case MARKUP_AREA :
 	case MARKUP_BODY :
-	case MARKUP_BR :
-	case MARKUP_DOCTYPE :
-	case MARKUP_EMBED :
 	case MARKUP_ERROR :
 	case MARKUP_FILE :
 	case MARKUP_HEAD :
-	case MARKUP_HR :
 	case MARKUP_HTML :
-	case MARKUP_IMG :
-	case MARKUP_INPUT :
-	case MARKUP_META :
 	case MARKUP_NONE :
 	case MARKUP_TITLE :
+
+	case MARKUP_APPLET :
+	case MARKUP_AREA :
+	case MARKUP_BR :
         case MARKUP_COMMENT :
+	case MARKUP_DOCTYPE :
+	case MARKUP_EMBED :
+	case MARKUP_HR :
+	case MARKUP_IMG :
+	case MARKUP_INPUT :
+	case MARKUP_ISINDEX :
+	case MARKUP_LINK :
+	case MARKUP_META :
+	case MARKUP_NOBR :
+	case MARKUP_SPACER :
+	case MARKUP_WBR :
             break;
 
         case MARKUP_CENTER :
@@ -707,8 +742,32 @@ add_link(uchar *name,		/* I - Name of link */
 
   if ((temp = find_link(name)) != NULL)
     temp->filename = filename;
-  else if (num_links < MAX_LINKS)
+  else
   {
+    // See if we need to allocate memory for links...
+    if (num_links >= alloc_links)
+    {
+      // Allocate more links...
+      alloc_links += ALLOC_LINKS;
+
+      if (num_links == 0)
+        temp = (link_t *)malloc(sizeof(link_t) * alloc_links);
+      else
+        temp = (link_t *)realloc(links, sizeof(link_t) * alloc_links);
+
+      if (temp == NULL)
+      {
+	progress_error(HD_ERROR_OUT_OF_MEMORY,
+	               "Unable to allocate memory for %d links - %s",
+	               alloc_links, strerror(errno));
+        alloc_links -= ALLOC_LINKS;
+	return;
+      }
+
+      links = temp;
+    }
+
+    // Add a new link...
     temp = links + num_links;
     num_links ++;
 
@@ -717,8 +776,7 @@ add_link(uchar *name,		/* I - Name of link */
     temp->filename = filename;
 
     if (num_links > 1)
-      qsort(links, num_links, sizeof(link_t),
-            (int (*)(const void *, const void *))compare_links);
+      qsort(links, num_links, sizeof(link_t), (compare_func_t)compare_links);
   }
 }
 
@@ -744,7 +802,7 @@ find_link(uchar *name)		/* I - Name to find */
   strncpy((char *)key.name, (char *)target, sizeof(key.name) - 1);
   key.name[sizeof(key.name) - 1] = '\0';
   match = (link_t *)bsearch(&key, links, num_links, sizeof(link_t),
-                            (int (*)(const void *, const void *))compare_links);
+                            (compare_func_t)compare_links);
 
   return (match);
 }
@@ -883,5 +941,5 @@ update_links(tree_t *t,		/* I - Document tree */
 
 
 /*
- * End of "$Id: html.cxx,v 1.17.2.10 2001/03/08 17:58:30 mike Exp $".
+ * End of "$Id: html.cxx,v 1.17.2.20 2001/10/20 21:49:16 mike Exp $".
  */
