@@ -1,5 +1,5 @@
 /*
- * "$Id: htmllib.cxx,v 1.41.2.18 2001/03/06 15:40:33 mike Exp $"
+ * "$Id: htmllib.cxx,v 1.41.2.41 2001/11/01 16:15:15 mike Exp $"
  *
  *   HTML parsing routines for HTMLDOC, a HTML document processing program.
  *
@@ -33,6 +33,7 @@
  *   get_text()          - Get all text from the given tree.
  *   htmlGetText()       - Get all text from the given tree.
  *   htmlGetMeta()       - Get document "meta" data...
+ *   htmlGetStyle()      - Get a style value from a node's STYLE attribute.
  *   htmlGetVariable()   - Get a variable value from a markup entry.
  *   htmlSetVariable()   - Set a variable for a markup entry.
  *   htmlSetBaseSize()   - Set the font sizes and spacings...
@@ -41,6 +42,8 @@
  *   compare_variables() - Compare two markup variables.
  *   compare_markups()   - Compare two markup strings...
  *   delete_node()       - Free all memory associated with a node...
+ *   insert_space()      - Insert a whitespace character before the
+ *                         specified node.
  *   parse_markup()      - Parse a markup string.
  *   parse_variable()    - Parse a markup variable string.
  *   compute_size()      - Compute the width and height of a tree entry.
@@ -54,7 +57,6 @@
  */
 
 #include "htmldoc.h"
-#include "string.h"
 #include <ctype.h>
 
 
@@ -62,7 +64,7 @@
  * Markup strings...
  */
 
-char		*_htmlMarkups[] =
+const char	*_htmlMarkups[] =
 		{
 		  "",		/* MARKUP_NONE */
 		  "!--",	/* MARKUP_COMMENT */
@@ -205,10 +207,15 @@ const char	*_htmlFonts[4][4] =
  * Local functions.
  */
 
+extern "C" {
+typedef int	(*compare_func_t)(const void *, const void *);
+}
+
 static int	write_file(tree_t *t, FILE *fp, int col);
 static int	compare_variables(var_t *v0, var_t *v1);
 static int	compare_markups(uchar **m0, uchar **m1);
 static void	delete_node(tree_t *t);
+static void	insert_space(tree_t *parent, tree_t *t);
 static int	parse_markup(tree_t *t, FILE *fp);
 static int	parse_variable(tree_t *t, FILE *fp);
 static int	compute_size(tree_t *t);
@@ -241,9 +248,9 @@ static uchar	indent[255] = "";
  */
 
 tree_t *			/* O - Pointer to top of file tree */
-htmlReadFile(tree_t *parent,	/* I - Parent tree entry */
-             FILE   *fp,	/* I - File pointer */
-	     char   *base)	/* I - Base directory for file */
+htmlReadFile(tree_t     *parent,/* I - Parent tree entry */
+             FILE       *fp,	/* I - File pointer */
+	     const char *base)	/* I - Base directory for file */
 {
   int		ch;		/* Character from file */
   uchar		*ptr,		/* Pointer in string */
@@ -307,7 +314,8 @@ htmlReadFile(tree_t *parent,	/* I - Parent tree entry */
     if (t == NULL)
     {
 #ifndef DEBUG
-      progress_error("Unable to allocate memory for HTML tree node!");
+      progress_error(HD_ERROR_OUT_OF_MEMORY,
+                     "Unable to allocate memory for HTML tree node!");
 #endif /* !DEBUG */
       break;
     }
@@ -357,169 +365,193 @@ htmlReadFile(tree_t *parent,	/* I - Parent tree entry */
       pos = ftell(fp) - 1;
 
       ch = getc(fp);
-      if (ch == ' ')
+      if (isspace(ch) || ch == '=' || ch == '<')
       {
        /*
-        * Illegal lone "<"!  Ignore it...
+        * Sigh...  "<" followed by anything but an element name is
+	* invalid HTML, but so many people have asked for this to
+	* be supported that we have added this hack...
 	*/
 
-	delete_node(t);
-	continue;
-      }
-      
-      if (ch != '/')
-        ungetc(ch, fp);
+	ptr = s;
+	if (have_whitespace)
+	{
+          *ptr++ = ' ';
+	  have_whitespace = 0;
+	}
 
-      if (parse_markup(t, fp) == MARKUP_ERROR)
-      {
-#ifndef DEBUG
-        progress_error("Unable to parse HTML element at %d!", pos);
-#endif /* !DEBUG */
+        *ptr++ = '<';
+	if (ch == '=')
+	  *ptr++ = '=';
+	else if (ch == '<')
+	  ungetc(ch, fp);
+	else
+	  have_whitespace = 1;
 
-        delete_node(t);
-        break;
-      }
+	*ptr++ = '\0';
 
-     /*
-      * Eliminate extra whitespace...
-      */
-
-      if (issuper(t->markup) || isblock(t->markup) ||
-          islist(t->markup) || islentry(t->markup) ||
-          istable(t->markup) || istentry(t->markup) ||
-	  t->markup == MARKUP_TITLE)
-        have_whitespace = 0;
-
-     /*
-      * If this is the matching close mark, or if we are starting the same
-      * markup, or if we've completed a list, we're done!
-      */
-
-      if (ch == '/')
-      {
-       /*
-        * Close markup; find matching markup...
-        */
-
-        for (temp = parent; temp != NULL; temp = temp->parent)
-          if (temp->markup == t->markup)
-            break;
-	  else if (temp->markup == MARKUP_EMBED)
-	  {
-	    temp = NULL;
-            break;
-	  }
-      }
-      else if (t->markup == MARKUP_EMBED)
-      {
-       /*
-        * Close any text blocks...
-	*/
-
-        for (temp = parent; temp != NULL; temp = temp->parent)
-          if (isblock(temp->markup) || islentry(temp->markup))
-            break;
-	  else if (istentry(temp->markup) || islist(temp->markup) ||
-	           issuper(temp->markup) || temp->markup == MARKUP_EMBED)
-	  {
-	    temp = NULL;
-	    break;
-	  }
-      }
-      else if (issuper(t->markup))
-      {
-        for (temp = parent; temp != NULL; temp = temp->parent)
-          if (issuper(temp->markup))
-            break;
-	  else if (istentry(temp->markup) || temp->markup == MARKUP_EMBED)
-	  {
-	    temp = NULL;
-            break;
-	  }
-      }
-      else if (islist(t->markup))
-      {
-        for (temp = parent; temp != NULL; temp = temp->parent)
-          if (isblock(temp->markup))
-	    break;
-	  else if (islentry(temp->markup) || istentry(temp->markup) ||
-	           issuper(temp->markup) || temp->markup == MARKUP_EMBED)
-	  {
-	    temp = NULL;
-            break;
-	  }
-
-      }
-      else if (islentry(t->markup))
-      {
-        for (temp = parent; temp != NULL; temp = temp->parent)
-          if (islentry(temp->markup) || isblock(temp->markup))
-            break;
-	  else if (islist(temp->markup) || issuper(temp->markup) ||
-	           istentry(temp->markup) || temp->markup == MARKUP_EMBED)
-          {
-	    temp = NULL;
-	    break;
-	  }
-      }
-      else if (isblock(t->markup))
-      {
-        for (temp = parent; temp != NULL; temp = temp->parent)
-          if (isblock(temp->markup) || islentry(temp->markup))
-            break;
-	  else if (istentry(temp->markup) || islist(temp->markup) ||
-	           issuper(temp->markup) || temp->markup == MARKUP_EMBED)
-	  {
-	    temp = NULL;
-	    break;
-	  }
-      }
-      else if (istable(t->markup))
-      {
-        for (temp = parent; temp != NULL; temp = temp->parent)
-          if (istable(temp->markup))
-	    break;
-	  else if (temp->markup == MARKUP_TABLE || temp->markup == MARKUP_EMBED)
-	  {
-	    temp = NULL;
-            break;
-	  }
-      }
-      else if (istentry(t->markup))
-      {
-        for (temp = parent; temp != NULL; temp = temp->parent)
-          if (istentry(temp->markup))
-            break;
-	  else if (temp->markup == MARKUP_TABLE || istable(temp->markup) ||
-	           temp->markup == MARKUP_EMBED)
-	  {
-	    temp = NULL;
-            break;
-	  }
+	t->markup = MARKUP_NONE;
+	t->data   = (uchar *)strdup((char *)s);
       }
       else
-        temp = NULL;
-
-      if (temp != NULL)
       {
-        DEBUG_printf(("%s>>>> Auto-ascend <<<\n", indent));
+       /*
+        * Start of a markup...
+	*/
 
-        delete_node(t);
+	if (ch != '/')
+          ungetc(ch, fp);
+
+	if (parse_markup(t, fp) == MARKUP_ERROR)
+	{
+  #ifndef DEBUG
+          progress_error(HD_ERROR_READ_ERROR,
+                         "Unable to parse HTML element at %d!", pos);
+  #endif /* !DEBUG */
+
+          delete_node(t);
+          break;
+	}
 
        /*
-        * If the markup doesn't start with a slash, or if it does but
-        * doesn't match up with the parent (i.e. <UL><LI>...<LI>...</UL>)
-        * then seek back so the parent entry gets a copy...
-        */
+	* Eliminate extra whitespace...
+	*/
 
-        if (ch != '/' || temp != parent)
-          fseek(fp, pos, SEEK_SET);	/* Make sure parent gets this markup... */
-        break;
-      }
-      else if (ch == '/')
-      {
-        delete_node(t);
-	continue;
+	if (issuper(t->markup) || isblock(t->markup) ||
+            islist(t->markup) || islentry(t->markup) ||
+            istable(t->markup) || istentry(t->markup) ||
+	    t->markup == MARKUP_TITLE)
+          have_whitespace = 0;
+
+       /*
+	* If this is the matching close mark, or if we are starting the same
+	* markup, or if we've completed a list, we're done!
+	*/
+
+	if (ch == '/')
+	{
+	 /*
+          * Close markup; find matching markup...
+          */
+
+          for (temp = parent; temp != NULL; temp = temp->parent)
+            if (temp->markup == t->markup)
+              break;
+	    else if (temp->markup == MARKUP_EMBED)
+	    {
+	      temp = NULL;
+              break;
+	    }
+	}
+	else if (t->markup == MARKUP_EMBED)
+	{
+	 /*
+          * Close any text blocks...
+	  */
+
+          for (temp = parent; temp != NULL; temp = temp->parent)
+            if (isblock(temp->markup) || islentry(temp->markup))
+              break;
+	    else if (istentry(temp->markup) || islist(temp->markup) ||
+	             issuper(temp->markup) || temp->markup == MARKUP_EMBED)
+	    {
+	      temp = NULL;
+	      break;
+	    }
+	}
+	else if (issuper(t->markup))
+	{
+          for (temp = parent; temp != NULL; temp = temp->parent)
+	    if (istentry(temp->markup) || temp->markup == MARKUP_EMBED)
+	    {
+	      temp = NULL;
+              break;
+	    }
+	}
+	else if (islist(t->markup))
+	{
+          for (temp = parent; temp != NULL; temp = temp->parent)
+            if (isblock(temp->markup))
+	      break;
+	    else if (islentry(temp->markup) || istentry(temp->markup) ||
+	             issuper(temp->markup) || temp->markup == MARKUP_EMBED)
+	    {
+	      temp = NULL;
+              break;
+	    }
+
+	}
+	else if (islentry(t->markup))
+	{
+          for (temp = parent; temp != NULL; temp = temp->parent)
+            if (islentry(temp->markup) || isblock(temp->markup))
+              break;
+	    else if (islist(temp->markup) || issuper(temp->markup) ||
+	             istentry(temp->markup) || temp->markup == MARKUP_EMBED)
+            {
+	      temp = NULL;
+	      break;
+	    }
+	}
+	else if (isblock(t->markup))
+	{
+          for (temp = parent; temp != NULL; temp = temp->parent)
+            if (isblock(temp->markup) || islentry(temp->markup))
+              break;
+	    else if (istentry(temp->markup) || islist(temp->markup) ||
+	             issuper(temp->markup) || temp->markup == MARKUP_EMBED)
+	    {
+	      temp = NULL;
+	      break;
+	    }
+	}
+	else if (istable(t->markup))
+	{
+          for (temp = parent; temp != NULL; temp = temp->parent)
+            if (istable(temp->markup))
+	      break;
+	    else if (temp->markup == MARKUP_TABLE || temp->markup == MARKUP_EMBED)
+	    {
+	      temp = NULL;
+              break;
+	    }
+	}
+	else if (istentry(t->markup))
+	{
+          for (temp = parent; temp != NULL; temp = temp->parent)
+            if (istentry(temp->markup))
+              break;
+	    else if (temp->markup == MARKUP_TABLE || istable(temp->markup) ||
+	             temp->markup == MARKUP_EMBED)
+	    {
+	      temp = NULL;
+              break;
+	    }
+	}
+	else
+          temp = NULL;
+
+	if (temp != NULL)
+	{
+          DEBUG_printf(("%s>>>> Auto-ascend <<<\n", indent));
+
+          delete_node(t);
+
+	 /*
+          * If the markup doesn't start with a slash, or if it does but
+          * doesn't match up with the parent (i.e. <UL><LI>...<LI>...</UL>)
+          * then seek back so the parent entry gets a copy...
+          */
+
+          if (ch != '/' || temp != parent)
+            fseek(fp, pos, SEEK_SET);	/* Make sure parent gets this markup... */
+          break;
+	}
+	else if (ch == '/')
+	{
+          delete_node(t);
+	  continue;
+	}
       }
     }
     else if (t->preformatted)
@@ -562,7 +594,7 @@ htmlReadFile(tree_t *parent,	/* I - Parent tree entry */
       t->markup = MARKUP_NONE;
       t->data   = (uchar *)strdup((char *)s);
 
-      DEBUG_printf(("%sfragment %s\n", indent, s));
+      DEBUG_printf(("%sfragment \"%s\"\n", indent, s));
     }
     else
     {
@@ -600,7 +632,7 @@ htmlReadFile(tree_t *parent,	/* I - Parent tree entry */
       }
 
       if (isspace(ch))
-        *ptr++ = ' ';
+        have_whitespace = 1;
 
       *ptr = '\0';
 
@@ -610,7 +642,7 @@ htmlReadFile(tree_t *parent,	/* I - Parent tree entry */
       t->markup = MARKUP_NONE;
       t->data   = (uchar *)strdup((char *)s);
 
-      DEBUG_printf(("%sfragment %s\n", indent, s));
+      DEBUG_printf(("%sfragment \"%s\"\n", indent, s));
     }
 
    /*
@@ -656,19 +688,34 @@ htmlReadFile(tree_t *parent,	/* I - Parent tree entry */
           if ((color = htmlGetVariable(t, (uchar *)"BGCOLOR")) != NULL &&
 	      !BodyColor[0])
 	    strcpy(BodyColor, (char *)color);
+
+          // Update the background image as necessary...
+          if ((filename = htmlGetVariable(t, (uchar *)"BACKGROUND")) != NULL)
+	    htmlSetVariable(t, (uchar *)"BACKGROUND",
+	                    (uchar *)fix_filename((char *)filename,
+			                          (char *)base));
+
+          htmlReadFile(t, fp, base);
           break;
 
       case MARKUP_IMG :
+          if (have_whitespace)
+	  {
+	    // Insert a space before this image...
+	    insert_space(parent, t);
+
+	    have_whitespace = 0;
+	  }
+
+          // Get the image alignment...
           t->valignment = ALIGN_TOP;
           get_alignment(t);
 
-         /*
-	  * Update the image source as necessary...
-	  */
-
+          // Update the image source as necessary...
           if ((filename = htmlGetVariable(t, (uchar *)"SRC")) != NULL)
-	    htmlSetVariable(t, (uchar *)"SRC",
-	                    (uchar *)fix_filename((char *)filename, base));
+	    htmlSetVariable(t, (uchar *)"REALSRC",
+	                    (uchar *)fix_filename((char *)filename,
+			                          (char *)base));
 
       case MARKUP_BR :
       case MARKUP_NONE :
@@ -695,7 +742,8 @@ htmlReadFile(tree_t *parent,	/* I - Parent tree entry */
           t->strikethrough = 0;
           t->preformatted  = 0;
           t->style         = STYLE_BOLD;
-          t->child         = htmlReadFile(t, fp, base);
+
+          htmlReadFile(t, fp, base);
           break;
 
       case MARKUP_P :
@@ -708,7 +756,8 @@ htmlReadFile(tree_t *parent,	/* I - Parent tree entry */
           t->superscript   = 0;
           t->strikethrough = 0;
           t->preformatted  = 0;
-          t->child         = htmlReadFile(t, fp, base);
+
+          htmlReadFile(t, fp, base);
           break;
 
       case MARKUP_PRE :
@@ -719,7 +768,8 @@ htmlReadFile(tree_t *parent,	/* I - Parent tree entry */
           t->superscript   = 0;
           t->strikethrough = 0;
           t->preformatted  = 1;
-          t->child         = htmlReadFile(t, fp, base);
+
+          htmlReadFile(t, fp, base);
           break;
 
       case MARKUP_BLOCKQUOTE :
@@ -730,12 +780,13 @@ htmlReadFile(tree_t *parent,	/* I - Parent tree entry */
       case MARKUP_DL :
           t->indent ++;
 
-          t->child = htmlReadFile(t, fp, base);
+          htmlReadFile(t, fp, base);
           break;
 
       case MARKUP_DIV :
           get_alignment(t);
-          t->child = htmlReadFile(t, fp, base);
+
+          htmlReadFile(t, fp, base);
           break;
 
       case MARKUP_HR :
@@ -756,18 +807,20 @@ htmlReadFile(tree_t *parent,	/* I - Parent tree entry */
       case MARKUP_EMBED :
           if ((filename = htmlGetVariable(t, (uchar *)"SRC")) != NULL)
 	  {
-	    filename = (uchar *)fix_filename((char *)filename, base);
+	    filename = (uchar *)fix_filename((char *)filename,
+	                                     (char *)base);
 
             if ((embed = fopen((char *)filename, "r")) != NULL)
             {
 	      strcpy(newbase, file_directory((char *)filename));
 
-              t->child = htmlReadFile(t, embed, newbase);
+              htmlReadFile(t, embed, newbase);
               fclose(embed);
             }
 #ifndef DEBUG
 	    else
-	      progress_error("Unable to embed \"%s\" - %s", filename,
+	      progress_error(HD_ERROR_FILE_NOT_FOUND,
+                             "Unable to embed \"%s\" - %s", filename,
 	                     strerror(errno));
 #endif /* !DEBUG */
 	  }
@@ -778,9 +831,12 @@ htmlReadFile(tree_t *parent,	/* I - Parent tree entry */
 	    t->halignment = t->parent->halignment;
 	  else
             t->halignment = ALIGN_CENTER;
+
           get_alignment(t);
+
           t->style = STYLE_BOLD;
-          t->child = htmlReadFile(t, fp, base);
+
+          htmlReadFile(t, fp, base);
           break;
 
       case MARKUP_TD :
@@ -788,9 +844,12 @@ htmlReadFile(tree_t *parent,	/* I - Parent tree entry */
 	    t->halignment = t->parent->halignment;
 	  else
             t->halignment = ALIGN_LEFT;
+
 	  get_alignment(t);
+
           t->style = STYLE_NORMAL;
-          t->child = htmlReadFile(t, fp, base);
+
+          htmlReadFile(t, fp, base);
           break;
 
       case MARKUP_FONT :
@@ -805,8 +864,18 @@ htmlReadFile(tree_t *parent,	/* I - Parent tree entry */
             else if (strstr((char *)face, "times") != NULL)
               t->typeface = TYPE_TIMES;
             else if (strstr((char *)face, "courier") != NULL)
+	    {
               t->typeface = TYPE_COURIER;
-            else if (strstr((char *)face, "symbol") != NULL)
+
+              if (have_whitespace)
+	      {
+		// Insert a space before monospaced text...
+		insert_space(parent, t);
+
+		have_whitespace = 0;
+	      }
+            }
+	    else if (strstr((char *)face, "symbol") != NULL)
               t->typeface = TYPE_SYMBOL;
           }
 
@@ -815,6 +884,14 @@ htmlReadFile(tree_t *parent,	/* I - Parent tree entry */
 
           if ((size = htmlGetVariable(t, (uchar *)"SIZE")) != NULL)
           {
+            if (have_whitespace)
+	    {
+	      // Insert a space before sized text...
+	      insert_space(parent, t);
+
+	      have_whitespace = 0;
+	    }
+
 	    if (isdigit(size[0]))
 	      sizeval = atoi((char *)size);
 	    else
@@ -828,37 +905,79 @@ htmlReadFile(tree_t *parent,	/* I - Parent tree entry */
               t->size = sizeval;
           }
 
-          t->child = htmlReadFile(t, fp, base);
+          htmlReadFile(t, fp, base);
           break;
 
       case MARKUP_BIG :
+          if (have_whitespace)
+	  {
+	    // Insert a space before big text...
+	    insert_space(parent, t);
+
+	    have_whitespace = 0;
+	  }
+
           if (t->size < 6)
             t->size += 2;
           else
             t->size = 7;
 
-          t->child = htmlReadFile(t, fp, base);
+          htmlReadFile(t, fp, base);
           break;
 
       case MARKUP_SMALL :
+          if (have_whitespace)
+	  {
+	    // Insert a space before small text...
+	    insert_space(parent, t);
+
+	    have_whitespace = 0;
+	  }
+
           if (t->size > 2)
             t->size -= 2;
           else
             t->size = 0;
 
-          t->child = htmlReadFile(t, fp, base);
+          htmlReadFile(t, fp, base);
           break;
 
       case MARKUP_SUP :
+          if (have_whitespace)
+	  {
+	    // Insert a space before superscript text...
+	    insert_space(parent, t);
+
+	    have_whitespace = 0;
+	  }
+
           t->superscript = 1;
-          t->size        = SIZE_SUP;
-          t->child       = htmlReadFile(t, fp, base);
+
+          if ((sizeval = t->size + SIZE_SUP) < 0)
+	    t->size = 0;
+	  else
+	    t->size = sizeval;
+
+          htmlReadFile(t, fp, base);
           break;
 
       case MARKUP_SUB :
+          if (have_whitespace)
+	  {
+	    // Insert a space before subscript text...
+	    insert_space(parent, t);
+
+	    have_whitespace = 0;
+	  }
+
           t->subscript = 1;
-          t->size      = SIZE_SUB;
-          t->child     = htmlReadFile(t, fp, base);
+
+          if ((sizeval = t->size + SIZE_SUB) < 0)
+	    t->size = 0;
+	  else
+	    t->size = sizeval;
+
+          htmlReadFile(t, fp, base);
           break;
 
       case MARKUP_KBD :
@@ -867,53 +986,85 @@ htmlReadFile(tree_t *parent,	/* I - Parent tree entry */
       case MARKUP_TT :
       case MARKUP_CODE :
       case MARKUP_SAMP :
+          if (have_whitespace)
+	  {
+	    // Insert a space before monospaced text...
+	    insert_space(parent, t);
+
+	    have_whitespace = 0;
+	  }
+
           t->typeface = TYPE_COURIER;
-          t->child    = htmlReadFile(t, fp, base);
+
+          htmlReadFile(t, fp, base);
           break;
 
       case MARKUP_B :
-          t->style |= STYLE_BOLD;
-          t->child = htmlReadFile(t, fp, base);
+          t->style = (style_t)(t->style | STYLE_BOLD);
+
+          htmlReadFile(t, fp, base);
           break;
 
       case MARKUP_DD :
           t->indent ++;
-          t->child = htmlReadFile(t, fp, base);
+
+          htmlReadFile(t, fp, base);
           break;
 
       case MARKUP_VAR :
-          t->style    |= STYLE_ITALIC;
+          t->style = (style_t)(t->style | STYLE_ITALIC);
       case MARKUP_DFN :
           t->typeface = TYPE_HELVETICA;
-          t->child    = htmlReadFile(t, fp, base);
+
+          htmlReadFile(t, fp, base);
           break;
 
       case MARKUP_STRONG :
-          t->style |= STYLE_BOLD;
+          t->style = (style_t)(t->style | STYLE_BOLD);
       case MARKUP_CITE :
       case MARKUP_DT :
       case MARKUP_EM :
       case MARKUP_I :
-          t->style |= STYLE_ITALIC;
-          t->child = htmlReadFile(t, fp, base);
+          t->style = (style_t)(t->style | STYLE_ITALIC);
+
+          htmlReadFile(t, fp, base);
           break;
 
       case MARKUP_U :
       case MARKUP_INS :
+          if (have_whitespace)
+	  {
+	    // Insert a space before underlined text...
+	    insert_space(parent, t);
+
+	    have_whitespace = 0;
+	  }
+
           t->underline = 1;
-          t->child     = htmlReadFile(t, fp, base);
+
+          htmlReadFile(t, fp, base);
           break;
 
       case MARKUP_STRIKE :
       case MARKUP_S :
       case MARKUP_DEL :
+          if (have_whitespace)
+	  {
+	    // Insert a space before struck-through text...
+	    insert_space(parent, t);
+
+	    have_whitespace = 0;
+	  }
+
           t->strikethrough = 1;
-          t->child         = htmlReadFile(t, fp, base);
+
+          htmlReadFile(t, fp, base);
           break;
 
       case MARKUP_CENTER :
           t->halignment = ALIGN_CENTER;
-          t->child      = htmlReadFile(t, fp, base);
+
+          htmlReadFile(t, fp, base);
           break;
 
       default :
@@ -923,7 +1074,7 @@ htmlReadFile(tree_t *parent,	/* I - Parent tree entry */
 
           get_alignment(t);
 
-          t->child = htmlReadFile(t, fp, base);
+          htmlReadFile(t, fp, base);
           break;
     }
   }  
@@ -1020,6 +1171,7 @@ write_file(tree_t *t,		/* I - Tree entry */
               putc('\n', fp);
               col = 0;
             }
+        default :
             break;
       }
 
@@ -1040,13 +1192,10 @@ write_file(tree_t *t,		/* I - Tree entry */
 
 	if (t->vars[i].value == NULL)
           col += fprintf(fp, "%s", t->vars[i].name);
-	else if (strchr((char *)t->vars[i].value, ' ') != NULL ||
-        	 strchr((char *)t->vars[i].value, '\t') != NULL ||
-        	 strchr((char *)t->vars[i].value, '\n') != NULL ||
-        	 strchr((char *)t->vars[i].value, '\r') != NULL)
-          col += fprintf(fp, "%s=\"%s\"", t->vars[i].name, t->vars[i].value);
+	else if (strchr((char *)t->vars[i].value, '\"') != NULL)
+          col += fprintf(fp, "%s=\'%s\'", t->vars[i].name, t->vars[i].value);
 	else
-          col += fprintf(fp, "%s=%s", t->vars[i].name, t->vars[i].value);
+          col += fprintf(fp, "%s=\"%s\"", t->vars[i].name, t->vars[i].value);
       }
 
       putc('>', fp);
@@ -1099,7 +1248,8 @@ write_file(tree_t *t,		/* I - Tree entry */
           case MARKUP_MENU :
               putc('\n', fp);
               col = 0;
-              break;
+          default :
+	      break;
         }
       }
     }
@@ -1193,7 +1343,7 @@ htmlDeleteTree(tree_t *parent)	/* I - Parent to delete */
 
 
 /*
- * 'htmlInsertTree()' - Insert a tree node to the parent.
+ * 'htmlInsertTree()' - Insert a tree node under the parent.
  */
 
 tree_t *			/* O - New entry */
@@ -1208,7 +1358,7 @@ htmlInsertTree(tree_t   *parent,/* I - Parent entry */
     return (NULL);
 
  /*
-  * Insert the tree entry to the end of the chain of children...
+  * Insert the tree entry at the beginning of the chain of children...
   */
 
   if (parent != NULL)
@@ -1369,16 +1519,16 @@ htmlNewTree(tree_t   *parent,	/* I - Parent entry */
 
     case MARKUP_SUP :
         t->superscript = 1;
-        t->size        = SIZE_SUP;
+        t->size        = SIZE_P + SIZE_SUP;
         break;
 
     case MARKUP_SUB :
         t->subscript = 1;
-        t->size      = SIZE_SUB;
+        t->size      = SIZE_P + SIZE_SUB;
         break;
 
     case MARKUP_B :
-        t->style |= STYLE_BOLD;
+        t->style = (style_t)(t->style | STYLE_BOLD);
         break;
 
     case MARKUP_DD :
@@ -1387,7 +1537,7 @@ htmlNewTree(tree_t   *parent,	/* I - Parent entry */
 
     case MARKUP_DT :
     case MARKUP_I :
-        t->style |= STYLE_ITALIC;
+        t->style = (style_t)(t->style | STYLE_ITALIC);
         break;
 
     case MARKUP_U :
@@ -1399,6 +1549,9 @@ htmlNewTree(tree_t   *parent,	/* I - Parent entry */
     case MARKUP_DEL :
         t->strikethrough = 1;
         break;
+
+    default :
+        break;
   }
 
   t->parent = parent;
@@ -1497,6 +1650,46 @@ htmlGetMeta(tree_t *tree,	/* I - Document tree */
 }
 
 
+/*
+ * 'htmlGetStyle()' - Get a style value from a node's STYLE attribute.
+ */
+
+uchar *				// O - Value or NULL
+htmlGetStyle(tree_t *t,		// I - Node
+             uchar  *name)	// I - Name (including ":")
+{
+  uchar		*ptr,		// Pointer in STYLE attribute
+		*bufptr;	// Pointer in buffer
+  int		ptrlen,		// Length of STYLE attribute
+		namelen;	// Length of name
+  static uchar	buffer[1024];	// Buffer for value
+
+
+  // See if we have a STYLE attribute...
+  if ((ptr = htmlGetVariable(t, (uchar *)"STYLE")) == NULL)
+    return (NULL);
+
+  // Loop through the STYLE attribute looking for the name...
+  for (namelen = strlen((char *)name), ptrlen = strlen((char *)ptr);
+       ptrlen > namelen;
+       ptr ++, ptrlen --)
+    if (strncasecmp((char *)name, (char *)ptr, namelen) == 0)
+    {
+      for (ptr += namelen; isspace(*ptr); ptr ++);
+
+      for (bufptr = buffer;
+           *ptr && *ptr != ';' && bufptr < (buffer + sizeof(buffer) - 1);
+	   *bufptr++ = *ptr++);
+
+      *bufptr = '\0';
+
+      return (buffer);
+    }
+
+  return (NULL);
+}
+
+
 /*
  * 'htmlGetVariable()' - Get a variable value from a markup entry.
  */
@@ -1515,7 +1708,7 @@ htmlGetVariable(tree_t *t,	/* I - Tree entry */
   key.name = name;
 
   v = (var_t *)bsearch(&key, t->vars, t->nvars, sizeof(var_t),
-                       (int (*)(const void *, const void *))compare_variables);
+                       (compare_func_t)compare_variables);
   if (v == NULL)
     return (NULL);
   else if (v->value == NULL)
@@ -1538,7 +1731,8 @@ htmlSetVariable(tree_t *t,	/* I - Tree entry */
 	key;			/* Search key */
 
 
-  DEBUG_printf(("%shtmlSetVariable(%08x, \"%s\", \"%s\")\n", indent, t, name, value));
+  DEBUG_printf(("%shtmlSetVariable(%08x, \"%s\", \"%s\")\n", indent, t, name,
+                value ? (const char *)value : "(null)"));
 
   if (t->nvars == 0)
     v = NULL;
@@ -1547,7 +1741,7 @@ htmlSetVariable(tree_t *t,	/* I - Tree entry */
     key.name = name;
 
     v = (var_t *)bsearch(&key, t->vars, t->nvars, sizeof(var_t),
-        	         (int (*)(const void *, const void *))compare_variables);
+        	         (compare_func_t)compare_variables);
   }
 
   if (v == NULL)
@@ -1581,7 +1775,7 @@ htmlSetVariable(tree_t *t,	/* I - Tree entry */
 
     if (t->nvars > 1)
       qsort(t->vars, t->nvars, sizeof(var_t),
-            (int (*)(const void *, const void *))compare_variables);
+            (compare_func_t)compare_variables);
   }
   else if (v->value != value)
   {
@@ -1656,7 +1850,8 @@ htmlSetCharSet(const char *cs)	/* I - Character set file to load */
     }
 #ifndef DEBUG
     else
-      progress_error("Unable to open psglyphs data file!");
+      progress_error(HD_ERROR_FILE_NOT_FOUND,
+                     "Unable to open psglyphs data file!");
 #endif /* !DEBUG */
   }
 
@@ -1674,7 +1869,8 @@ htmlSetCharSet(const char *cs)	/* I - Character set file to load */
     */
 
 #ifndef DEBUG
-    progress_error("Unable to open character set file %s!", cs);
+    progress_error(HD_ERROR_FILE_NOT_FOUND,
+                   "Unable to open character set file %s!", cs);
 #endif /* !DEBUG */
 
     for (i = 0; i < 256; i ++)
@@ -1748,7 +1944,8 @@ htmlSetCharSet(const char *cs)	/* I - Character set file to load */
       if ((fp = fopen(filename, "r")) == NULL)
       {
 #ifndef DEBUG
-        progress_error("Unable to open font width file %s!", _htmlFonts[i][j]);
+        progress_error(HD_ERROR_FILE_NOT_FOUND,
+                       "Unable to open font width file %s!", _htmlFonts[i][j]);
 #endif /* !DEBUG */
         continue;
       }
@@ -1861,6 +2058,62 @@ delete_node(tree_t *t)		/* I - Node to delete */
 }
 
 
+//
+// 'insert_space()' - Insert a whitespace character before the
+//                    specified node.
+//
+
+static void
+insert_space(tree_t *parent,	// I - Parent node
+             tree_t *t)		// I - Node to insert before
+{
+  tree_t	*space;		// Space node
+
+
+  // Allocate memory for the whitespace...
+  space = (tree_t *)calloc(sizeof(tree_t), 1);
+  if (space == NULL)
+  {
+#ifndef DEBUG
+    progress_error(HD_ERROR_OUT_OF_MEMORY,
+                   "Unable to allocate memory for HTML tree node!");
+#endif /* !DEBUG */
+    return;
+  }
+
+  // Set/copy font characteristics...
+  if (parent)
+  {
+    space->typeface = parent->typeface;
+    space->size     = parent->size;
+    space->style    = parent->style;
+  }
+  else
+  {
+    space->typeface = _htmlBodyFont;
+    space->size     = SIZE_P;
+  }
+
+  // Initialize element data...
+  space->markup = MARKUP_NONE;
+  space->data   = (uchar *)strdup(" ");
+
+  // Set tree pointers...
+  space->parent = parent;
+  space->prev   = t->prev;
+  space->next   = t;
+
+  if (space->prev)
+    space->prev->next = space;
+  else if (parent)
+    parent->child = space;
+
+  t->prev = space;
+
+  compute_size(space);
+}
+
+
 /*
  * 'parse_markup()' - Parse a markup string.
  */
@@ -1903,7 +2156,7 @@ parse_markup(tree_t *t,		/* I - Current tree entry */
   temp = (uchar **)bsearch(&mptr, _htmlMarkups,
                            sizeof(_htmlMarkups) / sizeof(_htmlMarkups[0]),
                            sizeof(_htmlMarkups[0]),
-                           (int (*)(const void *, const void *))compare_markups);
+                           (compare_func_t)compare_markups);
 
   if (temp == NULL)
   {
@@ -1919,7 +2172,7 @@ parse_markup(tree_t *t,		/* I - Current tree entry */
   }
   else
   {
-    t->markup = (markup_t)((char **)temp - _htmlMarkups);
+    t->markup = (markup_t)((const char **)temp - _htmlMarkups);
     cptr      = comment;
 
     DEBUG_printf(("%s%s\n", indent, markup));
@@ -2083,7 +2336,7 @@ compute_size(tree_t *t)		/* I - Tree entry */
     width_ptr  = htmlGetVariable(t, (uchar *)"WIDTH");
     height_ptr = htmlGetVariable(t, (uchar *)"HEIGHT");
 
-    img = image_load((char *)htmlGetVariable(t, (uchar *)"SRC"),
+    img = image_load((char *)htmlGetVariable(t, (uchar *)"REALSRC"),
                      _htmlGrayscale);
 
     if (width_ptr != NULL && height_ptr != NULL)
@@ -2235,7 +2488,10 @@ get_alignment(tree_t *t)	/* I - Tree entry */
   uchar	*align;			/* Alignment string */
 
 
-  if ((align = htmlGetVariable(t, (uchar *)"ALIGN")) != NULL)
+  if ((align = htmlGetVariable(t, (uchar *)"ALIGN")) == NULL)
+    align = htmlGetStyle(t, (uchar *)"text-align");
+
+  if (align != NULL)
   {
     if (strcasecmp((char *)align, "left") == 0)
       t->halignment = ALIGN_LEFT;
@@ -2253,7 +2509,10 @@ get_alignment(tree_t *t)	/* I - Tree entry */
       t->valignment = ALIGN_BOTTOM;
   }
 
-  if ((align = htmlGetVariable(t, (uchar *)"VALIGN")) != NULL)
+  if ((align = htmlGetVariable(t, (uchar *)"VALIGN")) == NULL)
+    align = htmlGetStyle(t, (uchar *)"vertical-align");
+
+  if (align != NULL)
   {
     if (strcasecmp((char *)align, "top") == 0)
       t->valignment = ALIGN_TOP;
@@ -2284,7 +2543,7 @@ fix_filename(char *filename,		/* I - Original filename */
   if (filename == NULL)
     return (NULL);
 
-  if (strcmp(base, ".") == 0 || strncmp(filename, "http://", 7) == 0)
+  if (strcmp(base, ".") == 0 || strstr(filename, "//") != NULL)
     return (file_find(Path, filename));
 
 #ifdef MAC
@@ -2389,5 +2648,5 @@ fix_filename(char *filename,		/* I - Original filename */
 
 
 /*
- * End of "$Id: htmllib.cxx,v 1.41.2.18 2001/03/06 15:40:33 mike Exp $".
+ * End of "$Id: htmllib.cxx,v 1.41.2.41 2001/11/01 16:15:15 mike Exp $".
  */
