@@ -1,5 +1,5 @@
 /*
- * "$Id: image.cxx,v 1.11.2.6 2001/02/27 02:13:13 mike Exp $"
+ * "$Id: image.cxx,v 1.11.2.15 2001/10/20 21:49:22 mike Exp $"
  *
  *   Image handling routines for HTMLDOC, a HTML document processing program.
  *
@@ -23,25 +23,26 @@
  *
  * Contents:
  *
- *   gif_read_cmap()     - Read the colormap from a GIF file...
- *   gif_get_block()     - Read a GIF data block...
- *   gif_get_code()      - Get a LZW code from the file...
- *   gif_read_image()    - Read a GIF image stream...
- *   gif_read_lzw()      - Read a byte from the LZW stream...
- *   image_compare()     - Compare two image filenames...
- *   image_copy()        - Copy image files to the destination directory...
- *   image_flush_cache() - Flush the image cache...
- *   image_getlist()     - Get the list of images that are loaded.
- *   image_load()        - Load an image file from disk...
- *   image_load_bmp()    - Read a BMP image file.
- *   image_load_gif()    - Load a GIF image file...
- *   image_load_jpeg()   - Load a JPEG image file.
- *   image_load_png()    - Load a PNG image file.
- *   image_need_mask()   - Allocate memory for the image mask...
- *   image_set_mask()  - Clear a bit in the image mask.
- *   read_word()         - Read a 16-bit unsigned integer.
- *   read_dword()        - Read a 32-bit unsigned integer.
- *   read_long()         - Read a 32-bit signed integer.
+ *   gif_read_cmap()      - Read the colormap from a GIF file...
+ *   gif_get_block()      - Read a GIF data block...
+ *   gif_get_code()       - Get a LZW code from the file...
+ *   gif_read_image()     - Read a GIF image stream...
+ *   gif_read_lzw()       - Read a byte from the LZW stream...
+ *   image_compare()      - Compare two image filenames...
+ *   image_copy()         - Copy image files to the destination directory...
+ *   image_flush_cache()  - Flush the image cache...
+ *   image_getlist()      - Get the list of images that are loaded.
+ *   image_load()         - Load an image file from disk...
+ *   image_load_bmp()     - Read a BMP image file.
+ *   image_load_gif()     - Load a GIF image file...
+ *   image_load_jpeg()    - Load a JPEG image file.
+ *   image_load_png()     - Load a PNG image file.
+ *   image_need_mask()    - Allocate memory for the image mask...
+ *   image_set_mask()     - Clear a bit in the image mask.
+ *   jpeg_error_handler() - Handle JPEG errors by not exiting.
+ *   read_word()          - Read a 16-bit unsigned integer.
+ *   read_dword()         - Read a 32-bit unsigned integer.
+ *   read_long()          - Read a 32-bit signed integer.
  */
 
 /*
@@ -83,8 +84,9 @@ typedef uchar	gif_cmap_t[256][3];
  * Local globals...
  */
 
-static int	num_images = 0;		/* Number of images in cache */
-static image_t	*images[MAX_IMAGES];	/* Images in cache */
+static int	num_images = 0,		/* Number of images in cache */
+		alloc_images = 0;	/* Allocated images */
+static image_t	**images = NULL;	/* Images in cache */
 static int	gif_eof = 0;		/* Did we hit EOF? */
 
 
@@ -101,13 +103,14 @@ static int	gif_read_image(FILE *fp, image_t *img, gif_cmap_t cmap,
 static int	gif_read_lzw(FILE *fp, int first_time, int input_code_size);
 
 static int	image_compare(image_t **img1, image_t **img2);
-static int	image_load_bmp(image_t *img, FILE *fp, int gray);
-static int	image_load_gif(image_t *img, FILE *fp, int gray);
-static int	image_load_jpeg(image_t *img, FILE *fp, int gray);
-static int	image_load_png(image_t *img, FILE *fp, int gray);
+static int	image_load_bmp(image_t *img, FILE *fp, int gray, int load_data);
+static int	image_load_gif(image_t *img, FILE *fp, int gray, int load_data);
+static int	image_load_jpeg(image_t *img, FILE *fp, int gray, int load_data);
+static int	image_load_png(image_t *img, FILE *fp, int gray, int load_data);
 static void	image_need_mask(image_t *img);
 static void	image_set_mask(image_t *img, int x, int y);
 
+static void		jpeg_error_handler(j_common_ptr);
 static int		read_long(FILE *fp);
 static unsigned short	read_word(FILE *fp);
 static unsigned int	read_dword(FILE *fp);
@@ -131,7 +134,11 @@ gif_read_cmap(FILE       *fp,		/* I  - File to read from */
   */
 
   if (fread(cmap, 3, ncolors, fp) < (size_t)ncolors)
+  {
+    progress_error(HD_ERROR_READ_ERROR,
+                   "Unable to read GIF colormap: %s", strerror(errno));
     return (-1);
+  }
 
  /*
   * Check to see if the colormap is a grayscale ramp...
@@ -184,6 +191,9 @@ gif_get_block(FILE  *fp,	/* I - File to read from */
     gif_eof = 1;
   else if (fread(buf, 1, count, fp) < (size_t)count)
   {
+    progress_error(HD_ERROR_READ_ERROR,
+                   "Unable to read GIF block of %d bytes: %s", count,
+                   strerror(errno));
     gif_eof = 1;
     return (-1);
   }
@@ -239,7 +249,11 @@ gif_get_code(FILE *fp,		/* I - File to read from */
     */
 
     if (done)
+    {
+      progress_error(HD_ERROR_READ_ERROR,
+                     "Not enough data left to read GIF compression code.");
       return (-1);	/* Sorry, no more... */
+    }
 
    /*
     * Move last two bytes to front of buffer...
@@ -530,11 +544,11 @@ static int			/* O - Result of comparison */
 image_compare(image_t **img1,	/* I - First image */
               image_t **img2)	/* I - Second image */
 {
-#if defined(WIN32) || defined(__EMX__)
+#ifdef WIN32
   return (strcasecmp((*img1)->filename, (*img2)->filename));
 #else
   return (strcmp((*img1)->filename, (*img2)->filename));
-#endif /* WIN32 || __EMX__ */
+#endif /* WIN32 */
 }
 
 
@@ -593,7 +607,8 @@ image_copy(const char *filename,/* I - Source file */
  */
 
 image_t *			/* O - Pointer to image */
-image_find(const char *filename)/* I - Name of image file */
+image_find(const char *filename,/* I - Name of image file */
+           int        load_data)/* I - 1 = load image data */
 {
   image_t	key,		/* Search key... */
 		*keyptr,	/* Pointer to search key... */
@@ -622,7 +637,12 @@ image_find(const char *filename)/* I - Name of image file */
     match = (image_t **)bsearch(&keyptr, images, num_images, sizeof(image_t *),
                                 (int (*)(const void *, const void *))image_compare);
     if (match != NULL)
-      return (*match);
+    {
+      if (load_data && !(*match)->pixels)
+        return (image_load((*match)->filename, (*match)->depth == 1, 1));
+      else
+        return (*match);
+    }
   }
 
   return (NULL);
@@ -648,10 +668,19 @@ image_flush_cache(void)
     if (images[i]->mask)
       free(images[i]->mask);
 
-    free(images[i]->pixels);
+    if (images[i]->pixels)
+      free(images[i]->pixels);
+
     free(images[i]);
   }
 
+  if (alloc_images)
+  {
+    free(images);
+
+    alloc_images = 0;
+  }
+
   num_images = 0;
 }
 
@@ -674,14 +703,16 @@ image_getlist(image_t ***ptrs)	/* O - Pointer to images array */
 
 image_t *			/* O - Pointer to image */
 image_load(const char *filename,/* I - Name of image file */
-           int        gray)	/* I - 0 = color, 1 = grayscale */
+           int        gray,	/* I - 0 = color, 1 = grayscale */
+           int        load_data)/* I - 1 = load image data, 0 = just info */
 {
   FILE		*fp;		/* File pointer */
   uchar		header[16];	/* First 16 bytes of file */
   image_t	*img,		/* New image buffer */
 		key,		/* Search key... */
 		*keyptr,	/* Pointer to search key... */
-		**match;	/* Matching image */
+		**match,	/* Matching image */
+		**temp;		/* Temporary array pointer */
   int		status;		/* Status of load... */
   const char	*realname;	/* Real filename */
 
@@ -707,12 +738,14 @@ image_load(const char *filename,/* I - Name of image file */
 
     match = (image_t **)bsearch(&keyptr, images, num_images, sizeof(image_t *),
                                 (int (*)(const void *, const void *))image_compare);
-    if (match != NULL)
+    if (match != NULL && (!load_data || (*match)->pixels))
     {
       (*match)->use ++;
       return (*match);
     }
   }
+  else
+    match = NULL;
 
  /*
   * Figure out the file type...
@@ -720,60 +753,87 @@ image_load(const char *filename,/* I - Name of image file */
 
   if ((realname = file_find(Path, filename)) == NULL)
   {
-    progress_error("Unable to find image file \"%s\"!", filename);
+    progress_error(HD_ERROR_FILE_NOT_FOUND,
+                   "Unable to find image file \"%s\"!", filename);
     return (NULL);
   }
 
   if ((fp = fopen(realname, "rb")) == NULL)
   {
-    progress_error("Unable to read image file \"%s\"!", filename);
+    progress_error(HD_ERROR_FILE_NOT_FOUND,
+                   "Unable to open image file \"%s\" (%s) for reading!",
+		   filename, realname);
     return (NULL);
   }
 
   if (fread(header, 1, sizeof(header), fp) == 0)
   {
-    progress_error("Unable to read image file \"%s\"!", filename);
+    progress_error(HD_ERROR_READ_ERROR,
+                   "Unable to read image file \"%s\"!", filename);
     fclose(fp);
     return (NULL);
   }
 
   rewind(fp);
 
- /*
-  * Allocate memory...
-  */
-
-  img = (image_t *)calloc(sizeof(image_t), 1);
-
-  if (img == NULL)
+  // See if the images array needs to be resized...
+  if (!match)
   {
-    progress_error("Unable to allocate memory for \"%s\"!", filename);
-    fclose(fp);
-    return (NULL);
+    if (num_images >= alloc_images)
+    {
+      // Yes...
+      alloc_images += ALLOC_FILES;
+
+      if (num_images == 0)
+	temp = (image_t **)malloc(sizeof(image_t *) * alloc_images);
+      else
+	temp = (image_t **)realloc(images, sizeof(image_t *) * alloc_images);
+
+      if (temp == NULL)
+      {
+	progress_error(HD_ERROR_OUT_OF_MEMORY,
+	               "Unable to allocate memory for %d images - %s",
+                       alloc_images, strerror(errno));
+	fclose(fp);
+	return (NULL);
+      }
+
+      images = temp;
+    }
+
+    // Allocate memory...
+    img = (image_t *)calloc(sizeof(image_t), 1);
+
+    if (img == NULL)
+    {
+      progress_error(HD_ERROR_READ_ERROR, "Unable to allocate memory for \"%s\"",
+                     filename);
+      fclose(fp);
+      return (NULL);
+    }
+
+    images[num_images] = img;
+
+    strcpy(img->filename, filename);
+    img->use = 1;
   }
+  else
+    img = *match;
 
-  images[num_images] = img;
-
-  strcpy(img->filename, filename);
-  img->use = 1;
-
- /*
-  * Load the image as appropriate...
-  */
-
+  // Load the image as appropriate...
   if (memcmp(header, "GIF87a", 6) == 0 ||
       memcmp(header, "GIF89a", 6) == 0)
-    status = image_load_gif(img,  fp, gray);
+    status = image_load_gif(img,  fp, gray, load_data);
   else if (memcmp(header, "BM", 2) == 0)
-    status = image_load_bmp(img, fp, gray);
+    status = image_load_bmp(img, fp, gray, load_data);
   else if (memcmp(header, "\211PNG", 4) == 0)
-    status = image_load_png(img, fp, gray);
+    status = image_load_png(img, fp, gray, load_data);
   else if (memcmp(header, "\377\330\377", 3) == 0 &&	/* Start-of-Image */
 	   header[3] >= 0xe0 && header[3] <= 0xef)	/* APPn */
-    status = image_load_jpeg(img, fp, gray);
+    status = image_load_jpeg(img, fp, gray, load_data);
   else
   {
-    progress_error("Unknown image file format for \"%s\"!", filename);
+    progress_error(HD_ERROR_BAD_FORMAT, "Unknown image file format for \"%s\"!", filename);
     fclose(fp);
     free(img);
     return (NULL);
@@ -783,15 +843,19 @@ image_load(const char *filename,/* I - Name of image file */
 
   if (status)
   {
-    progress_error("Unable to load image file \"%s\"!", filename);
-    free(img);
+    progress_error(HD_ERROR_READ_ERROR, "Unable to load image file \"%s\"!", filename);
+    if (!match)
+      free(img);
     return (NULL);
   }
 
-  num_images ++;
-  if (num_images > 1)
-    qsort(images, num_images, sizeof(image_t *),
-          (int (*)(const void *, const void *))image_compare);
+  if (!match)
+  {
+    num_images ++;
+    if (num_images > 1)
+      qsort(images, num_images, sizeof(image_t *),
+            (int (*)(const void *, const void *))image_compare);
+  }
 
   return (img);
 }
@@ -804,7 +868,8 @@ image_load(const char *filename,/* I - Name of image file */
 static int			/* O - 0 = success, -1 = fail */
 image_load_bmp(image_t *img,	/* I - Image to load into */
                FILE    *fp,	/* I - File to read from */
-	       int     gray)	/* I - Grayscale image? */
+	       int     gray,	/* I - Grayscale image? */
+               int     load_data)/* I - 1 = load image data, 0 = just info */
 {
   int		info_size,	/* Size of info header */
 		depth,		/* Depth of image (bits) */
@@ -854,6 +919,10 @@ image_load_bmp(image_t *img,	/* I - Image to load into */
 
   // Setup image and buffers...
   img->depth  = gray ? 1 : 3;
+
+  if (!load_data)
+    return (0);
+
   img->pixels = (uchar *)malloc(img->width * img->height * img->depth);
   if (img->pixels == NULL)
     return (-1);
@@ -871,6 +940,8 @@ image_load_bmp(image_t *img,	/* I - Image to load into */
   color = 0;
   count = 0;
   align = 0;
+  byte  = 0;
+  temp  = 0;
 
   for (y = img->height - 1; y >= 0; y --)
   {
@@ -1156,7 +1227,8 @@ image_load_bmp(image_t *img,	/* I - Image to load into */
 static int			/* O - 0 = success, -1 = fail */
 image_load_gif(image_t *img,	/* I - Image pointer */
                FILE    *fp,	/* I - File to load from */
-               int     gray)	/* I - 0 = color, 1 = grayscale */
+               int     gray,	/* I - 0 = color, 1 = grayscale */
+               int     load_data)/* I - 1 = load image data, 0 = just info */
 {
   uchar		buf[1024];	/* Input buffer */
   gif_cmap_t	cmap;		/* Colormap */
@@ -1244,6 +1316,9 @@ image_load_gif(image_t *img,	/* I - Image pointer */
           img->width  = (buf[5] << 8) | buf[4];
           img->height = (buf[7] << 8) | buf[6];
           img->depth  = gray ? 1 : 3;
+	  if (!load_data)
+	    return (0);
+
           img->pixels = (uchar *)malloc(img->width * img->height * img->depth);
           if (img->pixels == NULL)
             return (-1);
@@ -1261,21 +1336,25 @@ image_load_gif(image_t *img,	/* I - Image pointer */
 static int			/* O - 0 = success, -1 = fail */
 image_load_jpeg(image_t *img,	/* I - Image pointer */
                 FILE    *fp,	/* I - File to load from */
-                int     gray)	/* I - 0 = color, 1 = grayscale */
+                int     gray,	/* I - 0 = color, 1 = grayscale */
+                int     load_data)/* I - 1 = load image data, 0 = just info */
 {
   struct jpeg_decompress_struct	cinfo;		/* Decompressor info */
   struct jpeg_error_mgr		jerr;		/* Error handler info */
   JSAMPROW			row;		/* Sample row pointer */
 
 
-  cinfo.err = jpeg_std_error(&jerr);
+  jpeg_std_error(&jerr);
+  jerr.error_exit = jpeg_error_handler;
+
+  cinfo.err = &jerr;
   jpeg_create_decompress(&cinfo);
   jpeg_stdio_src(&cinfo, fp);
   jpeg_read_header(&cinfo, 1);
 
   cinfo.quantize_colors = 0;
 
-  if (gray)
+  if (gray || cinfo.num_components == 1)
   {
     cinfo.out_color_space      = JCS_GRAYSCALE;
     cinfo.out_color_components = 1;
@@ -1293,6 +1372,13 @@ image_load_jpeg(image_t *img,	/* I - Image pointer */
   img->width  = cinfo.output_width;
   img->height = cinfo.output_height;
   img->depth  = cinfo.output_components;
+
+  if (!load_data)
+  {
+    jpeg_destroy_decompress(&cinfo);
+    return (0);
+  }
+
   img->pixels = (uchar *)malloc(img->width * img->height * img->depth);
 
   if (img->pixels == NULL)
@@ -1325,7 +1411,8 @@ image_load_jpeg(image_t *img,	/* I - Image pointer */
 static int			/* O - 0 = success, -1 = fail */
 image_load_png(image_t *img,	/* I - Image pointer */
                FILE    *fp,	/* I - File to read from */
-               int     gray)	/* I - 0 = color, 1 = grayscale */
+               int     gray,	/* I - 0 = color, 1 = grayscale */
+               int     load_data)/* I - 1 = load image data, 0 = just info */
 {
   int		i;	/* Looping var */
   png_structp	pp;	/* PNG read pointer */
@@ -1366,6 +1453,13 @@ image_load_png(image_t *img,	/* I - Image pointer */
 
   img->width  = info->width;
   img->height = info->height;
+
+  if (!load_data)
+  {
+    png_read_destroy(pp, info, NULL);
+    return (0);
+  }
+
   img->pixels = (uchar *)malloc(img->width * img->height * 3);
 
   if (info->bit_depth < 8)
@@ -1487,8 +1581,8 @@ image_need_mask(image_t *img)	/* I - Image to add mask to */
 
 static void
 image_set_mask(image_t *img,	/* I - Image to operate on */
-                 int     x,	/* I - X coordinate */
-	         int     y)	/* I - Y coordinate */
+               int     x,	/* I - X coordinate */
+               int     y)	/* I - Y coordinate */
 {
   uchar		*maskptr;	/* Pointer into mask image */
   static uchar	masks[8] =	/* Masks for each bit */
@@ -1507,6 +1601,43 @@ image_set_mask(image_t *img,	/* I - Image to operate on */
 }
 
 
+/*
+ * 'image_unload()' - Unload an image from memory.
+ */
+
+void
+image_unload(image_t *img)	// I - Image
+{
+  if (!img)
+    return;
+
+  if (!img->use || !img->pixels)
+    return;
+
+  if (img->obj)
+    img->use = 0;
+  else
+    img->use --;
+
+  if (img->use)
+    return;
+
+  free(img->pixels);
+  img->pixels = NULL;
+}
+
+
+/*
+ * 'jpeg_error_handler()' - Handle JPEG errors by not exiting.
+ */
+
+static void
+jpeg_error_handler(j_common_ptr)
+{
+  return;
+}
+
+
 /*
  * 'read_word()' - Read a 16-bit unsigned integer.
  */
@@ -1560,5 +1691,5 @@ read_long(FILE *fp)               /* I - File to read from */
 
 
 /*
- * End of "$Id: image.cxx,v 1.11.2.6 2001/02/27 02:13:13 mike Exp $".
+ * End of "$Id: image.cxx,v 1.11.2.15 2001/10/20 21:49:22 mike Exp $".
  */
